int ttp_ppp_master, ttp_ppp_slave ;
int ttp_ppp_restart_flag ;

void relay_ppp(struct connection_state *cs, int cpid)
{
  fd_set readfds ;
  int done, maxfd, buflen, status, r ;
  char buf[8192] ;
  struct timeval timeout ;
  int trigger=0 ;
  
  fprintf(stderr,"relaying for ppp process %d...\n",cpid) ;

  if(ttp_ppp_master>cs->ttp_to_interface[0])
    maxfd=ttp_ppp_master ;
  else
    maxfd=cs->ttp_to_interface[0] ;
  for(done=0;!done;)
    {
      FD_ZERO(&readfds) ;
      FD_SET(ttp_ppp_master,&readfds) ;
      FD_SET(cs->ttp_to_interface[0],&readfds) ;
      timeout.tv_sec=32 ;
      timeout.tv_usec=0 ;

      r=select(maxfd+1,&readfds,NULL,NULL,&timeout) ;

      if(r>0)
	if(FD_ISSET(ttp_ppp_master,&readfds))
	  if((buflen=syswrap_read(ttp_ppp_master,buf,8192))>0)
	    {
	      sure_write(NULL,cs->interface_to_ttp[1],buf,buflen,0) ;
	      /* fprintf(stderr,"ttp_ppp: relay %d bytes from ppp to ttp\n",buflen) ; */
	    }

      if(r>0)
	if(FD_ISSET(cs->ttp_to_interface[0],&readfds))
	  if((buflen=syswrap_read(cs->ttp_to_interface[0],buf,8192))>0)
	    {
	      /* this means 1 byte is written at a time to ppp, in hope of solving
		 buffer stagnation problem, but it doesnt help... */
	      /* sure_write(NULL,ttp_ppp_master,buf,buflen,1) ; */
	      
	      fprintf(stderr,"*") ;
	      fflush(stderr) ;
	    
	      sure_write(NULL,ttp_ppp_master,buf,buflen,0) ;
	      /* fprintf(stderr,"ttp_ppp: relay %d bytes from ttp to ppp\n",buflen) ; */
	    }

      if(ttp_ppp_restart_flag)
	{
	  fprintf(stderr,"ttp_ppp: child has exited\n") ;
	  done=1 ;
	}
    }
  fprintf(stderr,"ttp_ppp: exit relay\n") ;
}

void ppp_child_handler(int flag)
{
  int cpid, status ;

  cpid=waitpid(-1,&status,WNOHANG) ;
  if(cpid>0)
    fprintf(stderr,"ttp_ppp: pppd (pid==%d) has exited\n",cpid) ;

  close(ttp_ppp_master) ;
  close(ttp_ppp_slave) ;

  ttp_ppp_restart_flag=1 ;
  signal(SIGCLD,ppp_child_handler) ;
}

void spawn_ppp(struct connection_state *cs, char *ipaddress)
{
  struct termios tio ;
  char *slavename ;
  int i, cpid, argc, length ;
  char command[1024] ;
  char ipparam[1024] ;
  char *argv[512] ;
  char *p ;

  for(;;)  /* if pppd exits, then restart it */
    {
      ttp_ppp_master=open("/dev/ptmx",O_RDWR) ;
      if(ttp_ppp_master<0)
	{
	  perror("ttp_ppp: cannot open /dev/ptmx") ;
	  exit(-1) ;
	}
      tcgetattr(ttp_ppp_master, &tio) ;
      cfmakeraw(&tio) ;
      grantpt(ttp_ppp_master) ;
      unlockpt(ttp_ppp_master) ;
      slavename=(char *)ptsname(ttp_ppp_master) ;
      if(slavename==NULL)
	{
	  perror("ttp_ppp: cannot get slave pty") ;
	  exit(-1) ;
	}
      fprintf(stderr,"opening slave pty %s for interface %s (master=%d)\n",slavename,cs->name,ttp_ppp_master) ;
      ttp_ppp_slave=open(slavename,O_RDWR) ;

      fprintf(stderr,"starting pppd (%s)...\n",pppdname) ;
      signal(SIGCLD,ppp_child_handler) ;
      ttp_ppp_restart_flag=0 ;
      cpid=fork() ;
      if(!cpid)
	{
	  int k ;
	  ioctl(0, TIOCNOTTY) ;
	  for(k=0;k<2;k++)
	    {
	      close(k) ;
	      dup(ttp_ppp_slave) ;
	    }
	  
	  /* give ttp a chance to get set up... */
	  sleep(16) ;
	  
	  /* !@#$ : remember to use ppp chap or pap authentication !!! */
	  sprintf(ipparam,"ttp") ;
	  p=strstr(pppdname,".") ;
	  if(p!=NULL)
	    strcat(ipparam,p) ;
	  /* sprintf(command,"%s local lcp-restart 2 lcp-max-failure 8192 lcp-echo-interval 8 lcp-echo-failure 64 mtu 8192 mru 8192 %s: nodefaultroute persist ipparam %s",pppdname,ipaddress,ipparam) ; */

	  sprintf(command,"%s local lcp-echo-interval 16 lcp-echo-failure 8 lcp-max-configure 8 mtu 2048 %s: nodefaultroute persist ipparam %s",pppdname,ipaddress,ipparam) ;
	  
	  fprintf(stderr,"executing command :\n\t%s\n",command) ;
	  
	  /* system(command) ; */
	  
	  argc=0 ;
	  length=strlen(command) ;
	  argv[0]=command ;
	  for(i=0;i<length+1;i++)
	    if((command[i]==' ') || (command[i]==0))
	      {
		command[i]=0 ;
		/* fprintf(stderr,"argv[%d] = %s\n",argc,argv[argc]) ; */
		argc++ ;
		argv[argc]=command+i+1 ;
	      }
	  argv[argc]=NULL ;
	  execvp(pppdname,argv) ;
	  exit(0) ;
	}
      fprintf(stderr,"new ttp_ppp pid = %d\n",cpid) ;
      relay_ppp(cs,cpid) ;
      fprintf(stderr,"ttp_ppp: relay done\n") ;
      close(ttp_ppp_slave) ;
      close(ttp_ppp_master) ;
    }
}

void init_ppp(struct connection_state *cs, char *ipaddress)
{
  pipe(cs->ttp_to_interface) ;
  pipe(cs->interface_to_ttp) ;

  cs->active=1 ;
  cs->pid=fork() ;
  if(!cs->pid)
    {
      spawn_ppp(cs,ipaddress) ;
      exit(0) ;
    }
  fprintf(stderr,"ppp interface spawned with pid %d\n",cs->pid) ;
}
