#define BUFSIZE 8192
#define STRLEN 8192

#define UDP_SEND_SOCKBUFSIZE 2048  /* socket buffer size */
#define UDP_RECEIVE_SOCKBUFSIZE 16384  /* socket buffer size */

struct connection_state *udp_cs=NULL ;

void udp_signal_handler(int flag)
{
  fprintf(stderr,"ttp_udp: interface %s is exiting after receiving SIGQUIT\n",udp_cs->name) ;
  exit(-1) ;
}

/* returns socket fd */
int init_udp_socket(char *hostname, int port, int local_port)
{
  int result= -1 ;
  int pb_socket, opt, handshake ;
  struct sockaddr_in locaddr, remaddr ;
  char temps[STRLEN] ;
  char message[STRLEN] ;
  char hostip[STRLEN] ;
  struct hostent *hostaddr ;

  if((pb_socket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP))<0)
    {
      perror("init_udp_socket: socket") ;
      if(errno==EMFILE)
	{
	  fprintf(stderr,"Too many open fds? What's wrong?...\n") ;
	  exit(-1) ;
	}
      return(-4) ;
    }

  opt = 1 ;
  if(setsockopt(pb_socket, SOL_SOCKET, SO_REUSEADDR,(char *) &opt, sizeof(opt))==-1)
    {
      perror("setsockopt (1)") ;
      close(pb_socket) ;
      return(-4) ;
    }

  opt = UDP_SEND_SOCKBUFSIZE ;
  if(setsockopt(pb_socket, SOL_SOCKET, SO_SNDBUF,(char *) &opt, sizeof(opt))==-1)
    {
      perror("setsockopt (2)") ;
      close(pb_socket) ;
      return(-4) ;
    }

  opt = UDP_RECEIVE_SOCKBUFSIZE ;
  if(setsockopt(pb_socket, SOL_SOCKET, SO_RCVBUF,(char *) &opt, sizeof(opt))==-1)
    {
      perror("setsockopt (3)") ;
      close(pb_socket) ;
      return(-4) ;
    }
      
  opt = STRLEN ;
  if(getsockopt(pb_socket,SOL_SOCKET,SO_SNDBUF,message,&opt)==-1)
    {
      perror("init_udp_socket: getsockopt (1)") ;
      close(pb_socket) ;
      return(-4) ;
    }

  opt=*((int *) message) ;

  opt = STRLEN ;
  if(getsockopt(pb_socket,SOL_SOCKET,SO_RCVBUF,message,&opt)==-1)
    {
      perror("init_udp_socket: getsockopt (2)") ;
      close(pb_socket) ;
      return(-4) ;
    }
  opt=*((int *) message) ;

  hostip[0]=0 ;

  if((remaddr.sin_addr.s_addr = inet_addr(hostname)) != -1)  /* user gave us an IP address */
    {
      sprintf(hostip,"%s",hostname) ;
    }
  else  /* resolve hostname */
    {
      hostaddr=gethostbyname(hostname) ;
      if(hostaddr!=NULL)
        {
          sprintf(hostip,"%d.%d.%d.%d",
                  (((unsigned char *) (hostaddr->h_addr_list[0]))[0]),
                  (((unsigned char *) (hostaddr->h_addr_list[0]))[1]),
                  (((unsigned char *) (hostaddr->h_addr_list[0]))[2]),
                  (((unsigned char *) (hostaddr->h_addr_list[0]))[3])) ;
        }
      else
        {
          int ipb0, ipb1, ipb2, ipb3 ;
          if(4==sscanf(hostname,"%d.%d.%d.%d",&ipb0,&ipb1,&ipb2,&ipb3))
            {
	      printf("init_udp_socket: cannot resolve host name, trying literal IP address\n") ;
              sprintf(hostip,"%d.%d.%d.%d",ipb0,ipb1,ipb2,ipb3) ;
            }
        }
      if(hostip[0]==0)
        {
	  fprintf(stderr,"init_udp_socket: cannot resolve host name : %s\n",hostname) ;
	  close(pb_socket) ;
          return(-2) ;      
        }

      /* if(inet_aton(hostip,&(remaddr.sin_addr))!=1) */
      if((remaddr.sin_addr.s_addr = inet_addr(hostip)) == -1)
        {
	  fprintf(stderr,"init_udp_socket: invalid host name\n") ;
	  close(pb_socket) ;
          return(-2) ;
        }
    }

  locaddr.sin_family = AF_INET ;
  locaddr.sin_addr.s_addr = htonl(INADDR_ANY) ;
  locaddr.sin_port = htons(local_port) ;

  if(bind(pb_socket,(struct sockaddr *) &locaddr,sizeof(locaddr))<0)
    {
      perror("ttp_udp: bind") ;
    }

  remaddr.sin_family = AF_INET ;
  remaddr.sin_port = htons(port) ;

  if(connect(pb_socket,(struct sockaddr*) &remaddr,sizeof(struct sockaddr_in))!=0)
    {
      sprintf(temps,"init_udp_socket: connect to %s:%d failed",hostname,port) ;
      perror(temps) ;
      close(pb_socket) ;
    }
  else
    {
      fprintf(stderr,"init_udp_socket: established connection to %s:%d\n",hostname,port) ;
      result=pb_socket ;
    }

  return(result) ;
}

void run_udp(struct connection_state *cs,char *remote_hostname, int remote_port, int local_port)
{
  int done ;
  int r, j, fd, maxfd, buflen, namelen ;
  fd_set readfds ;
  char buf[8192] ;
  struct timeval timeout ;
  struct sockaddr_in remaddr ;

  for(;;)
    {
      fprintf(stderr,"trying to connect to %s\n",cs->name) ;
      fd=init_udp_socket(remote_hostname, remote_port, local_port) ;
      if(fd>=0)
	{
	  fprintf(stderr,"udp interface %s becomes active\n",cs->name) ;
	  cs->active=1 ;
	  *((int *)buf)=htonl(TTP_ACTIVE) ;
	  sure_write(NULL,cs->interface_to_ttp[1],buf,sizeof(int),0) ;
	  
	  for(done=0;!done;)
	    {
	      FD_ZERO(&readfds) ;
	      maxfd=fd ;
	      FD_SET(fd,&readfds) ;
	      FD_SET(cs->ttp_to_interface[0],&readfds) ;
	      if(maxfd<cs->ttp_to_interface[0])
		maxfd=cs->ttp_to_interface[0] ;
	      timeout.tv_sec=64 ;
	      timeout.tv_usec=0 ;
	      
	      /* fprintf(stderr,"ttp_udp (%s): selecting...\n",cs->name) ; */
	      r=select(maxfd+1,&readfds,NULL,NULL,&timeout) ;

	      if(FD_ISSET(cs->ttp_to_interface[0],&readfds))
		if((buflen=syswrap_read(cs->ttp_to_interface[0],buf,8192))>0)
		  {
		    sure_write(NULL,fd,buf,buflen,0) ;  /* dont encrypt */
		    /* fprintf(stderr,"udp channel (%s) reads %d bytes from ttp and relays to udp\n",cs->name,buflen) ; */
		  }

	      if(FD_ISSET(fd,&readfds))
                {
		  buflen=syswrap_read(fd,buf,8192) ;
		  if(buflen>0)
		    {
		      /* relay buffer */
		      sure_write(NULL,cs->interface_to_ttp[1],buf,buflen,0) ;
		      /* fprintf(stderr,"udp channel (%s) reads %d bytes and relays to ttp\n",cs->name,buflen) ; */
		    }
		  else
		    {
		      fprintf(stderr,"ttp_udp: %s: selected, but %d bytes read\n",cs->name,buflen) ;
		      perror("read") ;
		      done=1 ;
		    }
                }
	    }
	  close(fd) ;
	  /* fprintf(stderr,"udp interface %s becomes inactive\n",cs->name) ; */
	  cs->active=0 ;
	  *((int *)buf)=htonl(TTP_INACTIVE) ;
	  sure_write(NULL,cs->interface_to_ttp[1],buf,sizeof(int),0) ;
	}
      sleep(16) ;
      fprintf(stderr,"ttp_udp: reinitializing interface: %s (mypid=%d)\n",cs->name,getpid()) ;
    }
}

void init_udp(struct connection_state *cs, char *remote_hostname, int remote_port, int local_port)
{
  int i ;

  pipe(cs->ttp_to_interface) ;
  pipe(cs->interface_to_ttp) ;

  cs->pid=fork() ;
  if(!cs->pid)
    {
      udp_cs=cs ;
      signal(SIGQUIT,udp_signal_handler) ;
      run_udp(cs, remote_hostname, remote_port, local_port) ;
    }
}
